import logging
from pathlib import Path
from typing import Iterator, List, Optional, Union

import cv2
import numpy
from PIL import Image

from RPA.core import geometry
from RPA.core.geometry import Region
from RPA.recognition.utils import to_image, clamp, log2lin


DEFAULT_CONFIDENCE = 80.0
LIMIT_FAILSAFE = 256

LOGGER = logging.getLogger(__name__)


class ImageNotFoundError(Exception):
    """Raised when template matching fails."""


def find(
    image: Union[Image.Image, Path],
    template: Union[Image.Image, Path],
    region: Optional[Region] = None,
    limit: Optional[int] = None,
    confidence: float = DEFAULT_CONFIDENCE,
) -> List[Region]:
    """Attempt to find the template from the given image.

    :param image:       Path to image or Image instance, used to search from
    :param template:    Path to image or Image instance, used to search with
    :param limit:       Limit returned results to maximum of `limit`.
    :param region:      Area to search from. Can speed up search significantly.
    :param confidence:  Confidence for matching, value between 1 and 100
    :return:            List of matching regions
    :raises ImageNotFoundError: No match was found
    """
    # Ensure images are in Pillow format
    image = to_image(image)
    template = to_image(template)

    # Convert confidence value to tolerance
    tolerance = _to_tolerance(confidence)

    # Crop image if requested
    if region is not None:
        region = geometry.to_region(region)
        image = image.crop(region.as_tuple())

    # Verify template still fits in image
    if template.size[0] > image.size[0] or template.size[1] > image.size[1]:
        raise ValueError("Template is larger than search region")

    # Do the actual search
    matches: List[Region] = []
    for match in _match_template(image, template, tolerance):
        matches.append(match)
        if limit is not None and len(matches) >= int(limit):
            break
        elif len(matches) >= LIMIT_FAILSAFE:
            LOGGER.warning("Reached maximum of %d matches", LIMIT_FAILSAFE)
            break

    if not matches:
        raise ImageNotFoundError("No matches for given template")

    # Convert region coördinates back to full-size coördinates
    if region is not None:
        matches = [match.move(region.left, region.top) for match in matches]

    return matches


def _to_tolerance(confidence):
    """Convert confidence value to tolerance.

    Confidence is a logarithmic scale from 1 to 100,
    tolerance is a linear scale from 0.01 to 1.00.
    """
    value = float(confidence)
    value = clamp(1, value, 100)
    value = log2lin(1, value, 100)
    value = value / 100.0
    return value


def _match_template(
    image: Image.Image, template: Image.Image, tolerance: float
) -> Iterator[Region]:
    """Use opencv's matchTemplate() to slide the `template` over
    `image` to calculate correlation coefficients, and then
    filter with a tolerance to find all relevant global maximums.
    """
    template_width, template_height = template.size

    if image.mode == "RGBA":
        image = image.convert("RGB")
    if template.mode == "RGBA":
        template = template.convert("RGB")

    image = numpy.array(image)
    template = numpy.array(template)

    # pylint: disable=no-member
    image = cv2.cvtColor(image, cv2.COLOR_RGB2BGR)
    template = cv2.cvtColor(template, cv2.COLOR_RGB2BGR)

    # Template matching result is a single channel array of shape:
    # Width:  Image width  - template width  + 1
    # Height: Image height - template height + 1
    coefficients = cv2.matchTemplate(image, template, cv2.TM_CCOEFF_NORMED)
    coeff_height, coeff_width = coefficients.shape

    while True:
        # The point (match_x, match_y) is the top-left of the best match
        _, match_coeff, _, (match_x, match_y) = cv2.minMaxLoc(coefficients)
        if match_coeff < tolerance:
            break

        # Zero out values for a template-sized region around the best match
        # to prevent duplicate matches for the same element.
        left = clamp(0, match_x - template_width // 2, coeff_width)
        top = clamp(0, match_y - template_height // 2, coeff_height)
        right = clamp(0, match_x + template_width // 2, coeff_width)
        bottom = clamp(0, match_y + template_height // 2, coeff_height)

        coefficients[top:bottom, left:right] = 0

        yield Region.from_size(match_x, match_y, template_width, template_height)



import logging
from collections import defaultdict
from difflib import SequenceMatcher
from pathlib import Path
from typing import Union, Dict, List, Generator, Optional

import pytesseract
from pytesseract import TesseractNotFoundError
from PIL import Image

from RPA.core import geometry
from RPA.core.geometry import Region
from RPA.recognition.utils import to_image, clamp

LOGGER = logging.getLogger(__name__)

# TODO: refer to conda package when created?
INSTALL_PROMPT = (
    "tesseract is not installed or not in PATH, "
    "see library documentation for installation instructions"
)

DEFAULT_CONFIDENCE = 80.0


def read(image: Union[Image.Image, Path]):
    """Scan image for text and return it as one string.

    :param image: Path to image or Image object
    """
    image = to_image(image)

    try:
        return pytesseract.image_to_string(image).strip()
    except TesseractNotFoundError as err:
        raise EnvironmentError(INSTALL_PROMPT) from err


def find(
    image: Union[Image.Image, Path],
    text: str,
    confidence: float = DEFAULT_CONFIDENCE,
    region: Optional[Region] = None,
    language: Optional[str] = None,
):
    """Scan image for text and return a list of regions
    that contain it (or something close to it).

    :param image: Path to image or Image object
    :param text: Text to find in image
    :param confidence: Minimum confidence for text similaritys
    :param region: Limit the region of the screen where to look for the text
    :param language: 3-character ISO 639-2 language code of the text.
     This is passed directly to the pytesseract lib in the lang parameter.
     See https://tesseract-ocr.github.io/tessdoc/Command-Line-Usage.html#using-one-language
    """  # noqa: E501
    image = to_image(image)
    confidence = clamp(1, float(confidence), 100)

    text = str(text).strip()
    if not text:
        raise ValueError("Empty search string")

    if region is not None:
        region = geometry.to_region(region)
        image = image.crop(region.as_tuple())

    try:
        data = pytesseract.image_to_data(
            image, lang=language, output_type=pytesseract.Output.DICT
        )
    except TesseractNotFoundError as err:
        raise EnvironmentError(INSTALL_PROMPT) from err

    lines = _dict_lines(data)
    matches = _match_lines(lines, text, confidence)

    if region is not None:
        for match in matches:
            match["region"] = match["region"].move(region.left, region.top)

    return matches


def _dict_lines(data: Dict) -> List:
    lines = defaultdict(list)
    for word in _iter_rows(data):
        if word["level"] != 5:
            continue

        if not word["text"].strip():
            continue

        key = "{:d}-{:d}-{:d}".format(
            word["block_num"], word["par_num"], word["line_num"]
        )
        region = Region.from_size(
            word["left"], word["top"], word["width"], word["height"]
        )

        # NOTE: Currently ignoring confidence in tesseract results
        lines[key].append({"text": word["text"], "region": region})
        assert len(lines[key]) == word["word_num"]

    return list(lines.values())


def _iter_rows(data: Dict) -> Generator:
    """Iterate dictionary of columns by row."""
    return (dict(zip(data.keys(), values)) for values in zip(*data.values()))


def _match_lines(lines: List[Dict], text: str, confidence: float) -> List[Dict]:
    """Find best matches between lines of text and target text,
    and return resulting bounding boxes and confidences.

    A line of N words will be matched to the given text in all 1 to N
    length sections, in every sequential position.
    """
    matches = []
    for line in lines:
        match = {}

        for window in range(1, len(line) + 1):
            for index in range(len(line) - window + 1):
                words = line[index : index + window]
                regions = [word["region"] for word in words]

                sentence = " ".join(word["text"] for word in words)
                ratio = SequenceMatcher(None, sentence, text).ratio() * 100.0

                if ratio < confidence:
                    continue

                if match and match["confidence"] >= ratio:
                    continue

                match = {
                    "text": sentence,
                    "region": Region.merge(regions),
                    "confidence": ratio,
                }

        if match:
            matches.append(match)

    return sorted(matches, key=lambda match: match["confidence"], reverse=True)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['RPA', 'RPA.recognition']

package_data = \
{'': ['*']}

install_requires = \
['numpy>=1.19.3,<2.0.0',
 'opencv-python-headless>=4.5.2,<4.6.0',
 'pillow>=9.1.1,<10.0.0',
 'pytesseract>=0.3.6,<0.4.0',
 'rpaframework-core>=10.4.0,<11.0.0']

setup_kwargs = {
    'name': 'rpaframework-recognition',
    'version': '5.1.0',
    'description': 'OCR capabilities used by RPA Framework',
    'long_description': 'rpaframework-recognition\n========================\n\nThis library enablous various recognition features with `RPA Framework`_\nlibraries, such as image template matching.\n\n.. _RPA Framework: https://rpaframework.org\n',
    'author': 'RPA Framework',
    'author_email': 'rpafw@robocorp.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://rpaframework.org/',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
